{

 The contents of this file are subject to the Mozilla Public License
 Version 1.1 (the "License"); you may not use this file except in compliance
 with the License. You may obtain a copy of the License at http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
 specific language governing rights and limitations under the License.

 The original code is SZCodeBaseX.pas, released 15. July, 2004.

 The initial developer of the original code is
 Sasa Zeman (public@szutils.net, www.szutils.net)

 Copyright(C) 2004-2005 Sasa Zeman. All Rights Reserved.

 Modified version released 2006 for WinBuilder by by booty#1
 http://www.boot-land.net/forums/index.php

}

unit OOCodeBaseX;

{$O+}// Optimization must be ON
{$R-}// Range checking must be OFF

interface

uses Windows, SysUtils, Types, Classes;

type
  TFastDecodeTable = array[0..255] of byte;

type
  TBaseDecoder = class
  protected
    BITS:      integer;
    BITS_D8:   integer;
    Buffer:    pByte;
    InputBufferBlockLen: integer;
    B8, I8:    integer;
    SZFastDecodeTable: TFastDecodeTable;
    OutStream: TStream;
    procedure SZUpdateFastDecodeTable(const Codes: string);
    function SZDecodeBaseXMemoryUpdate(pIN, pOUT: pByte; Size: integer): integer;
  public
    constructor Create(OutStream: TStream; BITS: integer; const Codes: string);
    destructor Destroy(); override;
    procedure Reset();
    procedure Process(Input: string);
  end;

  TBase64Decoder = class(TBasedecoder)
  public
    constructor Create(OutStream: TStream);
  end;

implementation

uses
  Math;

const
  SZCodes64 = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/';
  SZBITS64  = 6; // 64 = 2^6

constructor TBase64Decoder.Create(OutStream: TStream);
begin
  inherited Create(OutSTream, SZBITS64, SZCodes64);
end;

constructor TBaseDecoder.Create(OutStream: TStream; BITS: integer; const Codes: string);
var
  BufferLen: integer;
begin
  self.OutStream := OutStream;
  self.BITS      := BITS;
  self.BITS_D8   := BITS div 8;
  BufferLen      := 32 * 1024; // 32 KB output buffer length  
  GetMem(Buffer, BufferLen);
  InputBufferBlockLen := BufferLen div BITS;
  SZUpdateFastDecodeTable(Codes);
  Reset();
end;

destructor TBaseDecoder.Destroy();
begin
  FreeMem(Buffer);
  inherited;
end;

procedure TBaseDecoder.Reset();
begin
  B8 := 0;
  I8 := 0;
end;

procedure TBaseDecoder.SZUpdateFastDecodeTable(const Codes: string);
var
  i: integer;
begin
  FillChar(SZFastDecodeTable, 256, #0);
  for i := 1 to length(Codes) do
    SZFastDecodeTable[byte(Codes[i])] := i;
end;

procedure TBaseDecoder.Process(Input: string);
var
  LOut:      integer;
  pIN, pOUT: pByte;
  InputOffset: integer;
  InputLength: integer;
  RemainingBytes: integer;
  BlockLen:  integer;
begin
  InputLength := length(Input);
  RemainingBytes := InputLength;
  BlockLen := Min(InputBufferBlockLen, InputLength);
  InputOffset := 1;
  pOUT := Buffer;
  while (remainingBytes > 0) do
  begin
    pIN := @input[inputOffset];
    if (RemainingBytes < InputBufferBlockLen) then
      BlockLen := RemainingBytes;
    LOut := SZDecodeBaseXMemoryUpdate(pIN, pOUT, BlockLen);
    OutStream.Write(pOut^, LOut);
    RemainingBytes := RemainingBytes - BlockLen;
    inputOffset    := inputOffset + BlockLen;
  end;
end;


function TBaseDecoder.SZDecodeBaseXMemoryUpdate(pIN, pOUT: pByte;
  Size: integer): integer;
{
 Universal Decode algorithm for Base16, Base32 and Base64
 Reference: RFC 3548 - full compatibility
}

var
  i: integer;
  TotalIN, Count: integer;

begin
  TotalIN := Size;
  // Start decoding
  Count   := 0;
  for i := 1 to TotalIN do
  begin
    if SZFastDecodeTable[pIN^] > 0 then
    begin
      B8 := B8 shl BITS;
      B8 := B8 or (SZFastDecodeTable[pIN^] - 1);
      I8 := I8 + BITS;
      while I8 >= 8 do
      begin
        I8    := I8 - 8;
        pOUT^ := byte(B8 shr I8);
        Inc(pOUT);
        Inc(Count);
      end;
      Inc(pIN);
    end
    else if pIN^ = 13 then
      Inc(pIN)
    else if pIN^ = 10 then
      Inc(pIN)
    else
      break;
  end;
  Result := Count;
end;


end.
